/* ==============================================================================================================
 * 
 *                                VOIR_DEMAIN - Météo du lendemain sur Rennes
 *                                           released under GPLv3
 * 
 * --------------------------------------------------------------------------------------------------------------
 * Réalisé dans le cadre du hackathon Rudi des 9 et 10 décembre 2021, organisé par Les Petits Débrouillards Grand 
 * Ouest et L'Edulab de l'Université de Rennes 2, dans le cadre de l'appel à projet eRudi porté par Canopé.
 * --------------------------------------------------------------------------------------------------------------
 * Programme réalisé sur la base des exemples fournis par :
 * - Utilisation de la bibliothèque WiFiManager :
 *     . projetsdyi.fr :  https://projetsdiy.fr/librairie-wifimanager-connexion-wifi-projets-esp8266/
 *     . github de l'auteur (Tzapu) : https://github.com/tzapu/WiFiManager/tree/master/examples
 *  - Requête HTTPS depuis un D1 mini : 
 *     . Codebender : https://codebender.cc/sketch:322905#HTTPS%20Request%20Wemos%20D1.ino
 * - Bibliothèque JSON :
 *      . projetsdyi.fr : https://projetsdiy.fr/debuter-librairie-arduinojson-v6-librairie-arduino-esp32-esp8266-json/
 *      . calcul de la place nécessaire : https://arduinojson.org/v6/assistant/
 * --------------------------------------------------------------------------------------------------------------
 * Données météo extraites du site Météo Concept (gratuit jusqu'à 500 appels par jour) : https://api.meteo-concept.com/
 * Cela nécessite de s'inscrire sur le site afin d'obtenir un "token" qui sera indispensable lors des demandes de
 * prévision météo. Un token a déjà été obtenu pour notre maquette, mais en fonction du besoin (multiplication des
 * POCL "Voir Demain"), vous pourrez avoir à demander un nouveau "token" (via une nouvelle inscription sur le site), 
 * voire passer à une offre payante permettant d'envoyer plus de requêtes.
 * --------------------------------------------------------------------------------------------------------------
 * v0.1 - Baptiste Gaultier : mode simulation, animation des éléments et lumières du POCL.
 * v1.0 - Philippe Blusseau : gestion du Wi-Fi, accès au serveur météo et analyse des réponses du serveur.
 * =============================================================================================================*/

#include <WiFiManager.h>                    // Gestion de la connexion Wi-Fi (recherche de points d'accès)            
#include <WiFiClientSecure.h>               // Gestion de la connexion au serveur météo

#include <Adafruit_NeoPixel.h>              // Gestion des rubans de led multicouleurs
#include <Servo.h>                          // Gestion du servo-moteur

#define ARDUINOJSON_USE_DOUBLE 1         
#include <ArduinoJson.h>                    // Fonctions de décryptage des réponses du serveur météo

/* -------------------------------------------------------------------------------------------------------------
 *  Variables globales de l'application
 * ------------------------------------------------------------------------------------------------------------- */
 
WiFiManager GL_wifiManager;                               // Gestion de la connexion au Wi-Fi
const char* GL_ESP8266_AP_NAME= "POCL_Voir_Demain" ;      // Nom du point d'accès Wi-Fi du D1 Mini.
                                                          // - on n'utilisera pas de mot de passe.

const int NET_STATUS_NONE = 0;                            // Non connecté.
const int NET_STATUS_AP = 1;                              // Connexion en mode AP.
const int NET_STATUS_STANDARD = 2;                        // Connecté en mode standard.                                                         
int GL_networkStatus = NET_STATUS_NONE;                   // Etat initial                                                       
                                                         
const int PIN_POTENTIOMETER = A0;                         // PIN de lecture de la valeur du potentiomètre
const int PIN_SIMU_SWITCH = D0;                           // PIN de test de l'interrupteur de simu
const int PIN_SERVOMOTEUR = D3;                           // PIN de commande du servomoteur.
const int PIN_RGB_LEDS = D4;                              // PIN de commandes des leds RGB.

Servo GL_weatherServo;                                    // Servo pour l'aiguille indiquant la météo ...
const int GL_SERVO_NONE = -1 ;                            // ... position -1 on n'a pas (encore) reçu d'info Météo.
const int GL_SERVO_SUN = 0;                               // ... position pour le mode Soleil
const int GL_SERVO_CLOUDYSUN = 45;                        // ... position pour le mode Soleil voilé
const int GL_SERVO_CLOUDS = 90;                           // ... position pour le mode Nuages
const int GL_SERVO_RAIN = 135;                            // ... position pour le mode Orage
const int GL_SERVO_STORM = 180;                           // ... position pour le mode Nuages
int GL_servoPosition = GL_SERVO_NONE;                     // ... Position courante du servomoteur.

#define NB_WEATHER_DAYS 14
String GL_dateTime[NB_WEATHER_DAYS] ;                     // Dates & Heures des 14 prochains jours.
int    GL_weatherCode[NB_WEATHER_DAYS] ;                  // Météo (code précis) des 14 prochains jours.
int    GL_weatherAbstract[NB_WEATHER_DAYS] ;              // Météo (code concis entre 0 et 3) des 14 prochains jours.
int    GL_tomorrowAbstract = -1;                          // Le code concis de la météo de du lendemain (-1 = pas encore reçu d'info).

char*  GL_weatherText[] = { "Nuages & Soleil", "Nuages", "Soleil", "Orages" } ;   // Utilisé uniquement pour les traces

const int GL_NB_LEDS = 13;                              // Nombre de leds du ruban.

Adafruit_NeoPixel GL_weatherLeds                          // Gestion des leds RGB programmables.
    = Adafruit_NeoPixel(GL_NB_LEDS, PIN_RGB_LEDS, 
      NEO_GRB + NEO_KHZ800);

bool GL_realMode  ;                                       // false = mode simulation, true = on interrogera le serveur météo.
int  GL_potCode ;                                         // Code rettourné par le potentiomètre, entre 0 et 4.

unsigned long GL_REQUEST_INTERVAL_INIT = 2000 ;           // En mode réel, première interrogation du serveur météo au bout de 2 secondes.
unsigned long GL_REQUEST_INTERVAL = 7200000 ;             // En mode réel, consultation du serveur météo toutes les deux heures.
unsigned long GL_CHECK_INTERVAL = 500 ;                   // On consulte l'état de l'interrupteur et du potentiomètre toutes les 1/2 secondes
unsigned long GL_REPEAT_INTERVAL = 2000 ;                 // Réactivation de la séquence météo lumineuse toutes les 2 secondes.

unsigned long GL_nextServerCall ;                         // Prochain moment pour refaire une requête auprès du serveur météo
unsigned long GL_nextCheckCall ;                          // Prochain moment où consulter l'état de l'interrupteur et du potentiomètre.
unsigned long GL_nextRepeatCall ;                         // Prochain moment où réactiver la séquence météo lumineuse.

/* --------------------------------------------------------------------------------------------------------------
 *  STARTMYWIFI : Démarrage du Wi-Fi
 * ------------------------------------------------------------------------------------------------------------- */
void startMyWiFi() {
  
    /*
     * ... Clignotement de toutes les led en rouge pour indiquer le démarrage du Wi-Fi.
     */

    for (int i = 0; i < GL_NB_LEDS; i++) {
        GL_weatherLeds.setPixelColor(i, GL_weatherLeds.Color(/*Rouge*/ 255, /*Vert*/ 0, /* Bleu */ 0));
    }
    GL_weatherLeds.show(); 
  
    /* 
     * Mise en place de la connexion Wi-Fi. La librairie "WiFiManager.h" est très pratique, car elle permet d'enregistrer 
     * de façon pérenne (même si on débranche le D1 mini) les points d'accès Internet et leurs mots de passe. Si notre D1 mini
     * se trouve dans un nouvel environnement, il passe alors lui même en mode "point d'accès", et va proposer sur 
     * l'adresse 192.168.4.1 une interface permettant d'afficher les points d'accès environnants, sur lesquels on va pouvoir se 
     * connecter. Cela évite de mettre "en dur" dans le code le nom d'un point d'accès et du mot de passe associé.
     * A VERIFIER : TEST OPTION "NON BLOCKING"
     * Note : la librairie est bloquante : le programme ne continue qu'une fois connecté
     */ 

    Serial.println("Connexion au Wi-Fi ...");
    GL_wifiManager.setConfigPortalBlocking(false);              // Mode non bloquant
    if (GL_wifiManager.autoConnect(GL_ESP8266_AP_NAME)) {
      
          // Wi-Fi en mode standard ok --> on fait clignoter 5 fois les leds, cette fois-ci simultanément, et toujours en bleu

      GL_networkStatus = NET_STATUS_STANDARD ;
      
        int cBlue = 0 ;
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < GL_NB_LEDS; j++) {
                GL_weatherLeds.setPixelColor(j, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ cBlue));
            }
            GL_weatherLeds.show(); 
            cBlue = 255-cBlue ;     // on passe de 0 à 255, puis de 255 à 0, etc ...
            delay(100) ;
        }
    }
    else {

        Serial.println("Connexion Wi-Fi KO --> passage en mode AP.");     
        GL_networkStatus = NET_STATUS_AP;

    }

}

/* --------------------------------------------------------------------------------------------------------------
 *  SETUP : Initialisation
 * ------------------------------------------------------------------------------------------------------------- */
void setup() {
  
    /*
     * - Initialisation du port série (utilisé pour le débug)
     * - Allocation des entrées/sorties (interrupteur, potentiomètre, servomoteur)
     * - Initialisation des leds du ruban.
     */
    
    Serial.begin(115200);
    Serial.println("Démarrage de Voir Demain ...");

    pinMode(PIN_POTENTIOMETER, INPUT);  
    pinMode(PIN_SIMU_SWITCH, OUTPUT);  
    pinMode(PIN_SERVOMOTEUR, OUTPUT);  
    
    GL_weatherLeds.begin();
    GL_weatherLeds.setBrightness(50);

    /* 
     *  On fait clignoter toutes les leds en bleu à tour de rôle, pour signaler le démarrage
     */

    for (int i = 0; i < GL_NB_LEDS*5; i++) {
        GL_weatherLeds.setPixelColor(i%GL_NB_LEDS, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ 255));
        for (int j = 1; j < GL_NB_LEDS; j++) {
            GL_weatherLeds.setPixelColor((i+j)%GL_NB_LEDS, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ 0));
        }
        GL_weatherLeds.show(); 
        delay(40) ;
    }

    /*
     * Test de l'interrupteur (mode simu / mode réel) et du potentiomètre (code météo)
     */
      
    GL_realMode = checkSwitch() ;
    GL_potCode = checkPotCode() ;
    Serial.println(String("Init : Mode ") + String((GL_realMode ? "réel" : "simu")) + String(", Code Pot : ") + String(GL_potCode)) ;
      
    /* 
     *  Si mode réel, on lance le Wi-Fi ...
     */

    if (GL_realMode) {
          startMyWiFi() ;
    }

    /*
     * Activation des timers testés dans loop().
     */

    unsigned long myNow = millis() ;
    GL_nextServerCall = myNow + GL_REQUEST_INTERVAL_INIT ;
    GL_nextCheckCall = myNow + GL_CHECK_INTERVAL ;
    GL_nextRepeatCall = myNow + GL_REPEAT_INTERVAL ;

/* 
 *  JUSTE POUR VALIDATION DE LA TABLE DES CODES METEO
    testWeatherCodes() ;
 */
    
}

/* --------------------------------------------------------------------------------------------------------------
 *  LOOP : fonction appelée régulièrement par le système
 *  ------------------------------------------------------------------------------------------------------------- */
void loop() {

    /*
     * En mode non bloquant, on doit obligatoirement activer régulièrement le WiFimanager.
     */

    GL_wifiManager.process();

    /* 
     *  Interrogation de l'interrupteur et du potentiomètre toutes les GL_CHECK_INTERVAL millisecondes.
     *  S'il y a eu des changements depuis la fois précédente, on agit tout de suite
     */

    unsigned long myNow = millis() ;

    if (myNow >= GL_nextCheckCall) {

        bool myRealMode = checkSwitch() ;
        int myPotCode =  checkPotCode() ;

        if (myRealMode && !GL_realMode) {
            Serial.println("Chgt Simu --> Serveur Météo, code Pot : " + String(myPotCode)) ;
            if (WiFi.status() != WL_CONNECTED) {  
                Serial.println("Pas connecté --> Demande de connexion") ;            
                startMyWiFi() ;
                GL_nextServerCall = myNow + GL_REQUEST_INTERVAL_INIT ;
            }
        }
        else if (!myRealMode && GL_realMode) {
            Serial.println("Chgt Serveur Météo --> Simu, code Pot : " + String(myPotCode)) ;
        }

        if (myPotCode != GL_potCode) {
            Serial.println("Chgt Potentiomètre, " + String(GL_potCode) + " --> " + String(myPotCode)) ;
            GL_potCode = myPotCode ;
            showWeather(true) ;
            GL_nextRepeatCall = myNow + GL_REPEAT_INTERVAL ;
        }
      
        GL_realMode = myRealMode ;
        GL_nextCheckCall = myNow + GL_CHECK_INTERVAL ;

    }
        
    /*
     * Interrogation du serveur météo toutes les GL_REQUEST_INTERVAL millisecondes (si mode réel)
     */

    if (myNow >= GL_nextServerCall) {

        if (GL_realMode) {                                    
            if (WiFi.status() == WL_CONNECTED) {
                if (getServerResponse() == 1) {                         // Envoi de la requête et récupération de la réponse
                    applyServerResponse() ;                             // Traitement de la réponse.
                    showWeather(true) ;                                 // Application des nouveaux paramètres météo.
                }
                GL_nextServerCall = myNow + GL_REQUEST_INTERVAL ;
                GL_nextRepeatCall = myNow + GL_REPEAT_INTERVAL ;
            }
            else {
                startMyWiFi() ;
                GL_nextServerCall = myNow + GL_REQUEST_INTERVAL_INIT ;
                GL_nextRepeatCall = myNow + GL_REPEAT_INTERVAL ;
            }
        }
        
    }

    /*
     * Le reste du temps, on va continuer à animer la météo déjà analysée.
     */
     
    if (myNow >= GL_nextRepeatCall) {

        showWeather(false) ;
        GL_nextRepeatCall = myNow + GL_REPEAT_INTERVAL ;
        
    }

}
