/* ==============================================================================================================
 * 
 *                            VOIR_DEMAIN_SERVEUR - Traitement requête & réponse du serveur 
 * 
 * -------------------------------------------------------------------------------------------------------------
 * Comment ça marche ?
 * -------------------------------------------------------------------------------------------------------------
 * En fonction des serveurs "Open Data", l'interrogation peut se faire de différentes façons, et les données récupérées 
 * peuvent aussi avoir des formats différents. Cependant, bon nombre de ces serveurs sont basés sur le protocole
 * HTTP (on peut donc faire des demandes de données depuis un simple navigateur web), et sur le format de données JSON 
 * (que les navigateurs récents - en tout cas Firefox - comprennent également. En ce qui concerne la météo, cliquez sur 
 * (ou recopiez dans votre navigateur) le lien suivant ...
 * 
 *      https://api.meteo-concept.com/api/forecast/daily?token=54202d9b9f01cd1ee6d200372c7aec3661ac09f0ed9f526e85359de6eb064988&insee=35238
 * 
 * ... vous devriez avoir en retour un texte, que votre navigateur va quasi-immédiatement immédiatement reconnaître 
 * comme un format JSON. Voir onglet Voir_Demain_Codes(.ino) pour plus de détails. 
 * =============================================================================================================*/


    /*    
     * Serveur à consulter, port, et requête à envoyer à ce serveur.
     */

const char* GL_JSON_HOST = "api.meteo-concept.com";
const int   GL_HTTPS_PORT = 443;
const char* GL_JSON_REQUEST = "/api/forecast/daily?token=54202d9b9f01cd1ee6d200372c7aec3661ac09f0ed9f526e85359de6eb064988&insee=35238";  

    /* 
     *  La réponse sera entreposée ici. D'un point de vue général, par tatonnement, on voit qu'on ne peut pas dépasser 12000
     *  caractères, sinon l'ESP n'arrive pas à se connecter au serveur, voire plante carrément ! ... et ce nombre devra sans doute 
     *  être revu à la baisse si on ajoute encore des librairies, du code et des variables ... D'où l'intérêt de faire des requêtes 
     *  JSON précises donnant des réponses courtes ! Dans le cas du site Météo Concept, on n'a pas le choix, ce sont les prévisions 
     *  à 14 jours ou rien. On reçoit donc autour de 5300 caractères, on prévoit un peu plus large...
     */

const int GL_RESP_SIZE = 6000 ;
char GL_httpResponse[GL_RESP_SIZE] ; 

/* --------------------------------------------------------------------------------------------------------------
 *  GetServerResponse() : Envoi requête HTTP au serveur et récupération de la réponse
 *  valeur de retour : 1 = ok, 0 = pas de connexion serveur, -1 = pas de réponse, -2 = réponse anormale (trop de caractères)
 *  ------------------------------------------------------------------------------------------------------------- */
int getServerResponse() {
    
    const int API_TIMEOUT = 15000;      // Pour être sûr de recevoir l'en-tête de la réponse client.

    WiFiClientSecure myWiFiClient;

    /*
     * Comme la connexion est sécurisée (protocole HTTPS), il faudrait indiquer le certificat de la page que l'on recherche. 
     * Problème : dans tous les exemples trouvés, il est fait allusion à la récupération d'un certificat de type SHA-1 
     * sous Firefox, mais, d'une part le menu indiqué ... 
     *     Outils --> Informations sur la page --> Sécurité -> Afficher le certificat........empreinte numérique SHA-1
     * ... n'existe plus sous Firefox, et, d'autre part, l'algorithme SHA-1 n'est désormais plus soutenu car pas assez sécurisé ...
     * On va donc utiliser l'option magique ".setInsecure()". Pas trop grave dans notre exemple, où les données échangées
     * ne sont pas confidentielles.
     */

    myWiFiClient.setInsecure();
    myWiFiClient.setTimeout(API_TIMEOUT);

    /*
     * On fait clignoter les leds en bleu (5 flashs courts) pour signaler qu'on va chercher à se connecter au serveur
     */

    int cBlue = 0 ;
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < GL_NB_LEDS; j++) {
            GL_weatherLeds.setPixelColor(j, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ cBlue));
        }
        GL_weatherLeds.show(); 
        cBlue = 255-cBlue ;     // on passe de 0 à 255, puis de 255 à 0, etc ...
        delay(100) ;
    }

    /*
     * Connexion au serveur ... 
     */

    Serial.println("- Connexion au serveur [" + String(GL_JSON_HOST) + "]"); 
    int nbRetries = 6;
    while(!myWiFiClient.connect(GL_JSON_HOST, GL_HTTPS_PORT) && (nbRetries-- > 0)) {
        Serial.print(".");
        delay(1000);
    }   
    if(!myWiFiClient.connected()) {
        Serial.println("... Connexion impossible :-(");
        myWiFiClient.stop();
        return(0);
    }

    /*
     * Connecté ! --> On éteint les leds ...
     */

    GL_weatherLeds.clear(); 
    GL_weatherLeds.show(); 
    delay(200) ;  // Sinon on n'a pas le temps de voir les changements de couleur

    /*
     * Envoi de la requête URL
     */

    String myURL = String(GL_JSON_REQUEST);
    Serial.println("- Connexion OK --> Envoi requête URL - " + myURL);
    myWiFiClient.println("GET " + myURL + " HTTP/1.1") ;
    myWiFiClient.println("Host: " + String(GL_JSON_HOST)) ;
    myWiFiClient.println("Connection: close") ;
    myWiFiClient.println() ;
    
    /*
     * Attente de la réponse ....
     */
     
    int timeout = 5 * 10; // 5 seconds             
    while(!myWiFiClient.available() && (timeout-- > 0)){
        delay(100);
    }
    if(!myWiFiClient.available()) {
        Serial.println("... Pas de réponse :-(");
        myWiFiClient.stop();
        return(-1);
    }

    /*
     * Lecture de l'en-tête de la réponse (dont on ne fera rien)
     */
     
    Serial.println("- Réponse OK --> Lecture en-tête ...");
    while (myWiFiClient.connected()) {
        String line = myWiFiClient.readStringUntil('\n');
        if (line == "\r") {
            Serial.println("En-tête Lue --> Lecture corps de la réponse");
            break;
        }
    }

    /*
     * Entête reçue ! --> Lecture des données qui nous intéressent. On récupèrera caractère par caractère, 
     * (semble plus sûr que la fonction payload qui plante de temps en temps, d'après les forums. 
     * De plus, cela permet de faire défiler les leds pour montrer que ça suit son cours ...
     * On démarre donc le défilement avec la première led en bleu, les autres éteintes.
     */

    int numLed = 0; 
    GL_weatherLeds.setPixelColor(numLed, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ 255));
    for (int i = 1; i < GL_NB_LEDS; i++) {
         GL_weatherLeds.setPixelColor(i, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 0, /* Bleu */ 0));
    }
    GL_weatherLeds.show(); 

    /*
     * Pour une raison inconnue, quelque soit la méthode de lecture, il y a des caractères perturbateurs, qui apparaissent 
     * avant, au mileu, et après la structure JSON : si on ne fait rien, on obtient ceci (*) : 
     * 
     *      d6b[CR][LF]{"city":(...)[CR][LF]44e[CR][LF](...)}[CR][LF]0 
     *      
     * ... et en plus on ne sort pas de la boucle while" (?)... 
     * On va donc compter (variable myJSONLevel) les "{" et les "}" pour savoir où on en est, et sortir au dernier "}".
     * De plus, lorsqu'on tombe sur un [CR] en milieu de structure JSON, on attendra le prochain [CR] avant de reprendre en 
     * compte les caractères reçus. 
     * 
     * (*) [CR] = Carriage return (retour à la ligne), valeur décimale ascii 13
     *     [LF] = Line Feed (saut de ligne), , valeur décimale ascii 10
     */

    int myIndex = 0 ; 
    int myJSONLevel = 0 ;         // ie en dehors de la structure JSON.
    int myJSONLevelSaved = 0;     // pour retrouver à quel niveau on était avant de tomber sur la séquence perturbatrice...
  
    while(myWiFiClient.available() && myWiFiClient.connected() ){

        char myResp = myWiFiClient.read();

        if ((myResp < 32) || (myResp > 126)) {                        // Suppression caractères inutiles
            int i = myResp ; 
            // Serial.print("#SUP#") ;
            if (i == 13) {                                            // Donc, si on tombe sur un [CR] ...
                  if (myJSONLevel > 0) {                              // ... et qu'on est dans la structure ...
                      myJSONLevelSaved = myJSONLevel;                 // ... On sauvegarde notre niveau ....
                      myJSONLevel = -1;                               // ... et on le place temporairement à -1, ie "on attend le prochain [CR]".
                  }                                                   
                  else if (myJSONLevel == -1) {                      // et si on n'était plus dans la structure JSON ...
                      myJSONLevel = myJSONLevelSaved ;                // ... on y retourne avec le bon niveau d'accolades.
                  }
            }
            continue ;      
        }
          
        if (myResp == '{') { 

            myJSONLevel++ ; 

        } else if (myResp == '}') { 
          
            myJSONLevel-- ; 
            if ((myJSONLevel == 0) && (myIndex > 0)) {      // On sort de la structure JSON.
                GL_httpResponse[myIndex++] = myResp;        // ... ne pas oublier le dernier '}' avant de partir !
                GL_httpResponse[myIndex] = '\0';            // ... et caractère 'fin de chaine'. 
                break ;
            }
            
        }

        if (myJSONLevel > 0) {                              // On est dans la structure JSON.
                  
            GL_httpResponse[myIndex] = myResp;     
            if (myIndex++ >= GL_RESP_SIZE) {
                Serial.println("Réponse trop longue : " + String(GL_RESP_SIZE) + "caractères, et ne peut pas être traitée") ;
                myWiFiClient.stop();
                GL_weatherLeds.clear(); 
                GL_weatherLeds.show(); 
                return(-2);
            }
            GL_httpResponse[myIndex] = '\0';            // Vu sur forums : conseillé d'ajouté 'fin de chaîne' systématiquement ? 

            /*
             * Clignotement des leds en mode défilement : on éteint la led bleue, et on allume la suivante...
             */

            if (myIndex%50 == 0) { 
                GL_weatherLeds.setPixelColor(numLed, GL_weatherLeds.Color(0, 0, 0));
                numLed = (numLed + 1) % GL_NB_LEDS ;
                GL_weatherLeds.setPixelColor(numLed, GL_weatherLeds.Color(0, 0, 255));
                GL_weatherLeds.show(); 
            }
        }
        
        delay(1) ;    // Obligatoire ?, sinon plantage ...
        
    }

    /* Trace supprimée mais on garde au cas où ...

Serial.println();
Serial.println("-----");
Serial.println(String(GL_httpResponse)) ;
Serial.println("-----");

     */
            
    /*
     * Tout s'est bien passé --> on passe les leds en vert, et on se déconnecte du serveur.
     */
       
    for (int i = 0; i < GL_NB_LEDS; i++) {
        GL_weatherLeds.setPixelColor(i, GL_weatherLeds.Color(/*Rouge*/ 0, /*Vert*/ 255, /* Bleu */ 0));
    }
    GL_weatherLeds.show(); 
    delay(200) ;  // Sinon on n'a pas le temps de voir les changements de couleur

    myWiFiClient.stop();
    return(1) ;

}


/* --------------------------------------------------------------------------------------------------------------
 *  applyServerResponse() : Recupération des données JSON qui nous intéressent
 *  ------------------------------------------------------------------------------------------------------------- */
void applyServerResponse() {

    /*
     * le besoin en mémoire (capacity) pour l'analyse en mode json du texte reçu, doit être vérifié sur l'assistant 
     * https://arduinojson.org/v6/assistant/. On prend la réponse obtenue en lançant simplement la même requête HTTP 
     * dans un browser, on la copie dans cet assistant. Celui-ci nous indique alors que cette réponse (une fois supprimés
     * les caractères inutiles tels que blancs, tabulations, et sauts de ligne) fait 5296 caractères, soit plus d'1k,
     * et préconise alors d'utiliser 'DynamicJsonDocument'  et d'arrondir à 6144 octets.
     */

    Serial.println("applyServerResponse : début ...") ;
    DynamicJsonDocument myJSON(6144);

    /*
     * C'est là que ça devient intéressant : la fonction deserializeJson va transformer la réponse "texte" du serveur,
     * en une structure de données "myJSON", où il sera ensuite facile d'aller chercher les informations souhaitées.
     */

    Serial.println("applyServerResponse : lancement deserializeJson ...") ;
    DeserializationError error = deserializeJson(myJSON, GL_httpResponse, GL_RESP_SIZE);
    if (error) {
        Serial.println("deserializeJson() failed: " + String(error.f_str()));
        return;
    }
    
    /*
     * Nous pouvons maintenant extraire les informations que nous souhaitons. 
     * On enregistre les 14 jours de prévision dans les tableaux GL_dateTime, GL_weatherCode et GL_weatherAbstract,
     * même si cette maquette ne va utiliser que l'abstract des 5 prochains jours. 
     */

    Serial.println("applyServerResponse : city = " + String(myJSON["city"]["name"])) ;
    int myIndex = 0;

    for (JsonObject myRecord : myJSON["forecast"].as<JsonArray>()) {

        GL_dateTime[myIndex] = String(myRecord["datetime"]) ;
        GL_weatherCode[myIndex] = myRecord["weather"]; 
        GL_weatherAbstract[myIndex] = getWeatherAbstract(GL_weatherCode[myIndex]) ;
        
        // Trace qui pourra aussi être mise en commentaires
        const int myCode = GL_weatherCode[myIndex] ;
        const int myAbstract = GL_weatherAbstract[myIndex] ;
        Serial.print("[" + String(myIndex) + "] " + GL_dateTime[myIndex] + " / ") ;
        Serial.print(String(myCode) + " (" + GL_allWeatherCodes[myCode] + ")") ;
        Serial.println(" --> Code Météo " + String(myAbstract) + " (" + GL_weatherText[myAbstract] + ")") ;

        myIndex++ ;

    }
    
}
